/*
 *  MSSQL Communication Library -- TDS Handling
 *
 *  (c) 1998 Arno Pedusaar <psaar@fenar.ee>
 *  (c) 1998 Martin Mares <mj@ucw.cz>
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <netinet/in.h> /* for htons */
#include <time.h>

#include <endian.h>
#if !defined(__BYTE_ORDER) || !defined(__LITTLE_ENDIAN)
#error Invalid <endian.h>
#endif
#if __BYTE_ORDER==__LITTLE_ENDIAN
#define HW_LITTLE_ENDIAN
#elif __BYTE_ORDER==__BIG_ENDIAN
#define HW_BIG_ENDIAN
#else
#error Unknown endianity!
#endif

#include "tds.h"
#include "mssql.h"

#ifdef __GNUC__
#define NORET __attribute((noreturn))
#else
#define NORET
#endif

#define ERR handle_error

static void NORET
handle_error(char *msg, ...)
{
  va_list args;

  va_start(args, msg);
  fputs("mssql: ", stderr);
  vfprintf(stderr, msg, args);
  fputc('\n', stderr);
  exit(1);
}

static void *
mssql_malloc(unsigned int size)
{
void *x;

	x = malloc(size);
	if (!x)
		ERR("Out of memory (unable to allocate %d bytes).", size);
	return x;
}

static void
dump(unsigned char *data,int len,char *comment)
{
int pos;
int i;
unsigned char dd;

	pos=0;
	printf("%s\n",comment);

	while (pos<len) {
		if (pos % 16==0)
			printf("\n%04x  ",pos);

		for (i=0;i<16;++i)
			if (i+pos < len)
				printf("%02x " ,data[pos+i]);
			else
				printf("   ");

		for (i=0;i<16;++i)
			if (i+pos < len) {
				dd=data[pos+i];
				if (!iscntrl(dd))
					printf("%c",dd);
				else
					printf(".");
			}

		pos+=16;
	}

	printf("\n\n");
	fflush(stdout);
}

MSSQL *
allocMsSql(void)
{
MSSQL *s;

	s=(MSSQL *) mssql_malloc(sizeof(MSSQL));

	bzero(s,sizeof(MSSQL));

	s->fd=-1;
	s->blockSize=512;
	s->outBuf=(unsigned char *) mssql_malloc(s->blockSize);
	s->inBuf=(unsigned char *) mssql_malloc(s->blockSize);
	s->outFlag1=0x04;

	return s;
}


int
connectMsSql(MSSQL *s,char *name,int port)
{
	return s->connect(s,name,port);
}

int
disconnectMsSql(MSSQL *s)
{
	return s->disconnect(s);
}

static int
setFlag(MSSQL *s,unsigned char newFlag)
{
	s->outFlag1=newFlag;
	return 0;
}

static int
resizeMsSql(MSSQL *s,int newSize)
{
char *newBuf;
	
	if (s->outPos>newSize)
		return -1;

	newBuf=(char *) mssql_malloc(newSize);
	memcpy(newBuf,s->outBuf,(newSize<s->blockSize?newSize:s->blockSize));
	free(s->outBuf);
	s->outBuf=(unsigned char*)newBuf;

	newBuf=(char *) mssql_malloc(newSize);
	memcpy(newBuf,s->inBuf,(newSize<s->blockSize?newSize:s->blockSize));
	free(s->inBuf);
	s->inBuf=(unsigned char*)newBuf;

	s->blockSize=newSize;
	
	return 0;
}

/********************************************************
 *
 * Writers
 *
 ********************************************************/

/*
	Send current frame into network
 */
static int
doWrite(MSSQL *s,unsigned char lastFragment)
{
	s->outBuf[1]=lastFragment;
	*((unsigned short*) (s->outBuf+2))=htons(s->outPos);

	return s->write(s,s->outBuf,s->outPos);
}

/*
	Write len bytes from buf into output buffer,
	flush it when buffer reaches IO size
 */
static int
writeBuf(MSSQL *s,unsigned char *buf,int len)
{
int pos=0;
int spaceLeft;

	if (s->outPos==0) {
		s->outBuf[0]=s->outFlag1;
		s->outPos=8;
	}

	do {
		spaceLeft=s->blockSize-s->outPos;
		if (len>spaceLeft) {
			memcpy(s->outBuf+s->outPos,buf+pos,spaceLeft);
			s->outPos+=spaceLeft;
			pos+=spaceLeft;
			len-=spaceLeft;
			doWrite(s,0);
			s->outBuf[0]=s->outFlag1;
			s->outPos=8;
		} else {
			memcpy(s->outBuf+s->outPos,buf+pos,len);
			s->outPos+=len;
			break;
		}
	} while (len);

	return 0;
}

/*
	Mark current fragment as the last one and send it out
 */
static int
flushMsSql(MSSQL *s)
{
int rc=0;

	if (s->outPos>8) {
		rc=doWrite(s,1);
		s->outBuf[0]=s->outFlag1;
		s->outPos=8;
	} else
		ERR("Internal error #1");
	return rc;
}

/*
  Write string (as used in the login packet) of max. length len out:
  +-------------------------+--------------+
  | string body             |actual length |
  +-------------------------+--------------+
           len                    1
*/
static int
writeString(MSSQL *s,char *buf,int len)
{
char tmpBuf[len+1];
int stringLen;

	bzero(tmpBuf,len+1);

	if (buf) {
		stringLen=strlen(buf);
		stringLen=(stringLen>len?len:stringLen);
		memcpy(tmpBuf,buf,stringLen);
		tmpBuf[len]=stringLen;
	}
	
	return writeBuf(s,(unsigned char *)tmpBuf,len+1);
}

/*
	Write one char out
 */
static int
writeChar(MSSQL *s,unsigned char data)
{
	return writeBuf(s,&data,1);
}

#if 0

/*
	Write int8 out
 */
static int
write1Int(MSSQL *s,int data)
{
unsigned char buf;

	buf=data;
	return writeBuf(s,(unsigned char *)&buf,1);
}

/*
	Write int16 out
 */
static int
write2Int(MSSQL *s,int data)
{
short int tmpData;

	tmpData=data;
	return writeBuf(s,(unsigned char *)&tmpData,2);
}

#endif

/*
	Write int32 out
 */
static int
write4Int(MSSQL *s,int data)
{
	return writeBuf(s,(unsigned char *) &data,4);
}


/**********************************************************
 *
 * Readers
 *
 **********************************************************/

#define LAST_FRAGMENT 0x01

/*
	Read len bytes from current input stream into buf.
	Return:
		BUFFER_EMPTY - _last read_ exhausted the buffer completely.
		  (not very intuitive :-( )
		BUFFER_NOT_EMPTY - the read request was satisfied
		BUFFER_UNDERRUN - data requested was longer than that 
		  in input stream.
 */
static int
doRead(MSSQL *s,unsigned char *buf,int len)
{
int bufPos=0;

	while (len) {
		if (s->inFragmentSize==0) {
			if (s->read(s,s->inBuf,4))
				ERR("Read error");

			s->inFlag1=s->inBuf[0];
			s->inState=s->inBuf[1];
			s->inFragmentSize=(unsigned int)s->inBuf[2]<<8 |
			  (unsigned int)s->inBuf[3];

			if (s->blockSize<s->inFragmentSize)
				ERR("Block too short");

			if (s->read(s,s->inBuf+4,s->inFragmentSize-4))
				ERR("Read error");

			s->inPos=8;

			if (s->verbose>1)
				dump(s->inBuf,s->inFragmentSize,"");
		}

		if (s->inFragmentSize-s->inPos==0) {
			if (s->inState & LAST_FRAGMENT) {
				s->inState=0;
				s->inFragmentSize=0;
				return BUFFER_EMPTY;
			}
			s->inFragmentSize=0; /* rather hackish... */

			/* NB! Now goes to the top of the loop & reads the next fragment. */

		} else if (s->inFragmentSize-s->inPos>=len) {
			if (buf)
				memcpy(buf+bufPos,s->inBuf+s->inPos,len);
			s->inPos+=len;
			return BUFFER_NOT_EMPTY;
		} else {
			if (s->inState & LAST_FRAGMENT) {
				return BUFFER_UNDERRUN;
			}

			if (buf)
				memcpy(buf+bufPos,s->inBuf+s->inPos,s->inFragmentSize-s->inPos);

			len-=s->inFragmentSize-s->inPos;
			bufPos+=s->inFragmentSize-s->inPos;
			s->inFragmentSize=0;
		}
	}

	return BUFFER_NOT_EMPTY;
}

static int
read1Int(MSSQL *s,int *data)
{
unsigned char buf;
int rc;

	rc=doRead(s,&buf,1);
	if (data)
		*data=buf;
	return rc;
}

static int
read2Int(MSSQL *s,int *data)
{
short int tmpData;
int rc;

	rc=doRead(s,(unsigned char *)&tmpData,2);
	if (data)
		*data=(int)tmpData;
	return rc;
}

static int
read2UInt(MSSQL *s,int *data)
{
short unsigned int tmpData;
int rc;

	rc=doRead(s,(unsigned char *)&tmpData,2);
	if (data)
		*data=(int)tmpData;
	return rc;
}

static int
read4Int(MSSQL *s,int *data)
{
int tmpData;
int rc;

	rc=doRead(s,(unsigned char *)&tmpData,4);
	if (data)
		*data=(int)tmpData;
	return rc;
}

/*
	Read intN
	
	+--------------------+------------------+
	| data len [0,1,2,4] | data             |
	+--------------------+------------------+
	       1                [0,1,2,4]

	data len==0 ===> data==0
 */
static int
readNInt(MSSQL *s,int *data)
{
int len;
int rc;

	rc=read1Int(s,&len);

	switch (len) {
	case 0:
		if (data)
			*data=0;
		break;
	case 1:
		rc|=read1Int(s,data);
		break;
	case 2:
		rc|=read2Int(s,data);
		break;
	case 4:
		rc|=read4Int(s,data);
		break;
	default:
		ERR("Unknown NInt type");
	}

	return rc;
}

/*
	Read small string (0-255 symbols)
	+------------------+------------------+
	| data len 0...255 | data             |
	+------------------+------------------+
	       1                0...255
*/
static int
readSmallString(MSSQL *s,char *buf,int max)
{
int len;
int rc;

	rc=read1Int(s,&len);

	if (len>=max) {
		rc|=doRead(s,(unsigned char *)buf,max);
		if (buf)
			buf[max]=0;
		rc|=doRead(s,0,len-max);
	} else {
		rc|=doRead(s,(unsigned char *)buf,len);
		if (buf)
			buf[len]=0;
	}
	return rc;
}


/*
	Read medium string (0-65535??? symbols)
	+--------------------+------------------+
	| data len 0...65535 | data             |
	+--------------------+------------------+
	       2                0...65535???
*/
static int
readMediumString(MSSQL *s,char *buf,int max)
{
int len;
int rc;

	rc=read2UInt(s,&len);

	if (len>=max) {
		rc|=doRead(s,(unsigned char *)buf,max);
		if (buf)
			buf[max]=0;
		rc|=doRead(s,0,len-max);
	} else {
		rc|=doRead(s,(unsigned char *)buf,len);
		if (buf)
			buf[len]=0;
	}
	return rc;
}

/*
	Read single precision float
 */
static int
read4Float(MSSQL *s,double *buf)
{
float tmp;
int rc;

	rc=doRead(s,(unsigned char *)&tmp,4);
	if (buf)
		*buf=tmp;
	return rc;
}

/*
	Read double precision float.
 */
static int
read8Float(MSSQL *s,double *buf)
{
	return doRead(s,(unsigned char *)buf,8);
}

/*
	Read floatN
	
	+------------------+------------------+
	| data len [0,4,8] | data             |
	+------------------+------------------+
	       1                [0,4,8]

	data len==0 ===> data==0.0
 */
static int
readNFloat(MSSQL *s,double *buf)
{
int len;
int rc;

	rc=read1Int(s,&len);

	switch (len) {
	case 0:
		if (buf)
			*buf=0.0;
		break;
	case 4:
		rc|=read4Float(s,buf);
		break;
	case 8:
		rc|=read8Float(s,buf);
		break;
	default:
		ERR("Unknown NFloat type");
	}
	return rc;
}

/*
	Read money64
	
	Little endian protocol (quite funny...):
          MSW         LSW
	+-----------+-----------+
	| LSB...MSB | LSB...MSB |
	+-----------+-----------+
	   4           4

	Big endian protocol:
	+-----------+
	| MSB...LSB |
	+-----------+
	  8
*/
static int
read8Money(MSSQL *s,long long *buf)
{
#if defined(HW_LITTLE_ENDIAN)
unsigned char tmp[8];
int rc;

	rc=doRead(s,tmp,8);

	if (buf) {
		*(unsigned long long*)buf=
		  (unsigned long long)tmp[0]<<(4*8) |
		  (unsigned long long)tmp[1]<<(5*8) |
		  (unsigned long long)tmp[2]<<(6*8) |
		  (unsigned long long)tmp[3]<<(7*8) |
		  (unsigned long long)tmp[4] |
		  (unsigned long long)tmp[5]<<8 |
		  (unsigned long long)tmp[6]<<(2*8) |
		  (unsigned long long)tmp[7]<<(3*8);
	}
#elif defined(HW_BIG_ENDIAN)
int rc;

	rc=doRead(s,(unsigned char *) buf,8);
#endif

	return rc;
}

/*
	Read money32
	
	+-----------+
	| data      |
	+-----------+
	  8
*/
static int
read4Money(MSSQL *s,long long *buf)
{
int tmp;
int rc;

	rc=doRead(s,(unsigned char *)&tmp,4);

	if (buf)
		*buf=tmp;
	
	return rc;
}

/*
	Read moneyN
	
	+------------------+------------+
	| data len [0,4,8] | data       |
	+------------------+------------+
	  1                  [0,4,8]

	data len==0 ===> data==0.0
*/
static int
readNMoney(MSSQL *s,long long *buf)
{
int len;
int rc;

	rc=read1Int(s,&len);

	switch (len) {
	case 0:
		if (buf)
			*buf=0;
		break;
	case 4:
		rc|=read4Money(s,buf);
		break;
	case 8:
		rc|=read8Money(s,buf);
		break;
	default:
		ERR("Unknown NMoney type");
	}
	return rc;
}

/*
	Dunno...
 */
static int
readNumeric(MSSQL *s,double *buf)
{
int len;
int rc;
unsigned char tmp[64];

	rc=read1Int(s,&len);
	rc|=doRead(s,tmp,len);
	return rc;
}

#define MSSQL_EPOCH 25567 /* 1900/01/01 00:00:00 ---- days ---->  1970/01/01 00:00:00*/
#define SECONDS_PER_DAY 24*60*60

/*
	Read datetime4, return unix-style timestamp
	
	+----------------+--------------------+
	| days from 1900 | minutes from 00:00 |
	+----------------+--------------------+
	  2                2
 */
static int
read4Datetime(MSSQL *s,time_t *buf)
{
int rc;
unsigned int days;
unsigned int minutes;

	rc=read2UInt(s,&days);
	rc|=read2UInt(s,&minutes);
	if (buf)
		*buf=(days-MSSQL_EPOCH)*SECONDS_PER_DAY+minutes*60;
	return rc;
}

/*
	Read datetime8, return unix-style timestamp
	
	+----------------+--------------------------+
	| days from 1900 | 300th-seconds from 00:00 |
	+----------------+--------------------------+
	  4                4
 */
static int
read8Datetime(MSSQL *s,time_t *buf)
{
int rc;
int days;
int seconds;

	rc=read4Int(s,&days);
	rc|=read4Int(s,&seconds);
	if (buf)
		*buf=(days-MSSQL_EPOCH)*SECONDS_PER_DAY+seconds/300;
	return rc;
}

/*
	Read datetimeN
	
	+------------------+------------+
	| data len [0,4,8] | data       |
	+------------------+------------+
	  1                  [0,4,8]

	data len==0 ===> data==0.0
*/
static int
readNDatetime(MSSQL *s,time_t *buf)
{
int len;
int rc;

	rc=read1Int(s,&len);
	switch (len) {
	case 0:
		break;
	case 4:
		rc|=read4Datetime(s,buf);
		break;
	case 8:
		rc|=read8Datetime(s,buf);
		break;
	default:
		ERR("Unknown DateTime type");
	}

	return rc;
}

/****************************************************************
 *
 * Token handlers
 *
 ****************************************************************/
 
/*
	Free column metadata list
 */
static void
freeMetaData(MSSQL *s)
{
COLUMN *c,*next;

	for (c=s->columns.next;c;c=next) {
		next=c->next;
		free(c);
	}

	s->columns.next=0;
	s->columnCount=0;
}

/*
	Enumerate result column names
	
	+-----+-----+------+
	| tok | len | data |
	+-----+-----+------+
	 1      2     len

	data:
	  n * smallString, n - count of colums; not given explicitly,
	must be detected by data length. Bizzarre...
*/
static int
handleColumnNames(MSSQL *s)
{
int len;
int rc;
COLUMN *c;

	rc=read2Int(s,&len);

	freeMetaData(s);
	c=&(s->columns);

	while (len) {
		if (rc&BUFFER_UNDERRUN)
			return BUFFER_UNDERRUN;
		if (rc&BUFFER_EMPTY)
			return BUFFER_EMPTY;

		c->next=(COLUMN *)malloc(sizeof(COLUMN));
		bzero(c->next,sizeof(COLUMN));
		
		rc|=readSmallString(s,c->next->name,32);
		len-=strlen(c->next->name)+1;
		s->columnCount++;

		c=c->next;
	}

	return rc;
}

/*
	Enumerate result column metadata
	
	+-----+-----+------+
	| tok | len | data |
	+-----+-----+------+
	 1      2     len

	data:
	  n * colMeta, n - count of colums; not given explicitly,
	must be detected by data length. Bizzarre...

	colMeta:
	
	+--------+------+-----------------+
	| flags  | type | additional data |
	+--------+------+-----------------+
          4        1      0...x

	additional data:
 	  VARBINARY,INTN,VARCHAR,BINARY,CHAR,SENSITIVITY,
	  BOUNDARY,FLOATN,MONEYN,DATETIMEN:
		+-------------+
	  	| max. length |
	  	+-------------+
	  	  1

	  DECIMAL,NUMERIC:
		+-----+-----------+-------+
		| len | precision | scale |
		+-----+-----------+-------+
		  1     1           1
		  
	  LONGCHAR,BINARY:
		+--------------+
		| max. length? |
		+--------------+
		  4

	  IMAGE,TEXT:
		+--------------+--------------+
		| max. length? | name         |
		+--------------+--------------+
		  4              mediumString
 */
static int
handleMetaData(MSSQL *s)
{
int len;
int rc;
COLUMN *c;

	rc=read2Int(s,&len);
	c=s->columns.next;

	while (len) {
		if (c==0)
			ERR("TDS error! count(col names) != count(col meta)");
		if (rc&BUFFER_UNDERRUN)
			return BUFFER_UNDERRUN;
		if (rc&BUFFER_EMPTY)
			return BUFFER_EMPTY;

		rc|=read4Int(s,(int *)&(c->flags));
		rc|=read1Int(s,&(c->type));
		len-=5;

		switch (c->type) {
		case T_INT1:
		case T_BIT:
			c->len=1;
			break;
		case T_INT2:
			c->len=2;
			break;
		case T_INT4:
		case T_DATETIME4:
		case T_FLOAT4:
		case T_MONEY4:
			c->len=4;
			break;
		case T_MONEY8:
		case T_DATETIME8:
		case T_FLOAT8:
			c->len=8;
			break;
		case T_VARBINARY:
		case T_INTN:
		case T_VARCHAR:
		case T_BINARY:
		case T_CHAR:
		case T_SENSITIVITY:
		case T_BOUNDARY:
		case T_FLOATN:
		case T_MONEYN:
		case T_DATETIMEN:
			rc|=read1Int(s,&(c->len));
			len--;
			break;
		case T_DECIMAL:
		case T_NUMERIC:
			rc|=read1Int(s,&(c->len));
			rc|=read1Int(s,&(c->precision));
			rc|=read1Int(s,&(c->scale));
			len-=3;
			break;
		case T_LONGCHAR:
		case T_LONGBINARY:
			rc|=read4Int(s,&(c->len));
			len-=4;
			break;
		case T_IMAGE:
		case T_TEXT:
			rc|=read4Int(s,&(c->len));
			rc|=readMediumString(s,c->tableName,32);
			len-=4+strlen(c->tableName); /* AIAIAI!!!!!! PUUG!!! */
			break;
		default:
			ERR("Unknown type");
		}
		c=c->next;
	}

	if (c!=0)
		ERR("TDS error! count(col names) != count(col meta)");

	return rc;
}

/*
	Return (hardcoded) binding SQL type ---> C type
 */
static int
toCType(int dbType)
{
	switch (dbType) {
	case T_INT1:
	case T_BIT:
	case T_INT2:
	case T_INT4:
	case T_INTN:
		return T_C_INTEGER;
	case T_DATETIME4:
	case T_DATETIME8:
	case T_DATETIMEN:
		return T_C_TIME;
	case T_FLOAT4:
	case T_FLOAT8:
	case T_FLOATN:
		return T_C_DOUBLE;
	case T_MONEY4:
	case T_MONEY8:
	case T_MONEYN:
		return T_C_LONGLONG;
	case T_VARBINARY:
	case T_VARCHAR:
	case T_BINARY:
	case T_CHAR:
		return T_C_STRING;
	case T_SENSITIVITY:
	case T_BOUNDARY:
	case T_DECIMAL:
	case T_NUMERIC:
	case T_LONGCHAR:
	case T_LONGBINARY:
	case T_IMAGE:
	case T_TEXT:
		return T_C_UNDEFINED;
	default:
		ERR("toCType: Unknown data type %02x.\n",dbType);
		return 0;
	}
}

/*
	Result row
	
	+-----+------+
	| tok | data |
	+-----+------+
          1     n

	Simple concatenation, without field delimiters. 
	Must be read sequentially, as per meta data.
	Data length is _not_ explicitly given!
 */
static int
handleRowData(MSSQL *s)
{
int rc;
void *val;
COLUMN *c;
int iVal;
double dVal;
long long llVal;
char sVal[257];
time_t ttVal;
struct tm *t;

	for (c=s->columns.next,rc=0;c;c=c->next) {
		if (rc&BUFFER_UNDERRUN)
			return BUFFER_UNDERRUN;
		if (rc&BUFFER_EMPTY)
			return BUFFER_EMPTY;

		rc=0;
		val=c->val;

		if (val && (toCType(c->type)==T_C_UNDEFINED || 
		            toCType(c->type)!=c->boundType)) {
			fprintf(stderr,"C type mismatch!\n");
			val=0;
		}

		switch (c->type) {
		case T_INT1:
		case T_BIT:
			rc|=read1Int(s,&iVal);
			break;
		case T_INT2:
			rc|=read2Int(s,&iVal);
			break;
		case T_INT4:
			rc|=read4Int(s,&iVal);
			break;
		case T_INTN:
			rc|=readNInt(s,&iVal);
			break;
		case T_FLOAT4:
			rc|=read4Float(s,&dVal);
			break;
		case T_FLOAT8:
			rc|=read8Float(s,&dVal);
			break;
		case T_FLOATN:
			rc|=readNFloat(s,&dVal);
			break;
		case T_MONEY4:
			rc|=read4Money(s,&llVal);
			break;
		case T_MONEY8:
			rc|=read8Money(s,&llVal);
			break;
		case T_MONEYN:
			rc|=readNMoney(s,&llVal);
			break;
		case T_DATETIME4:
			rc|=read4Datetime(s,&ttVal);
			break;
		case T_DATETIME8:
			rc|=read8Datetime(s,&ttVal);
			break;
		case T_DATETIMEN:
			rc|=readNDatetime(s,&ttVal);
			break;
		case T_VARBINARY:
		case T_VARCHAR:
		case T_BINARY:
		case T_CHAR:
			rc|=readSmallString(s,sVal,256);
			break;
		case T_SENSITIVITY:
			fprintf(stderr,"SENSITIVITY: not supported!\n");
			break;
		case T_BOUNDARY:
			fprintf(stderr,"BOUNDARY: not supported!\n");
			break;
		case T_DECIMAL:
		case T_NUMERIC:
			rc|=readNumeric(s,0);
			break;
		case T_LONGCHAR:
		case T_LONGBINARY:
			fprintf(stderr,"LONGCHAR | LONGBINARY: not supported!\n");
			break;
		case T_IMAGE:
		case T_TEXT:
			fprintf(stderr,"IMAGE | TEXT: not supported!\n");
			break;
		default:
			ERR("Data type %02x not supported!\n", c->type);
		}

/*
	MySQLish interface...
 */
		switch (toCType(c->type)) {
		case T_C_INTEGER:
			sprintf(c->asString,"%d",iVal);
			break;
		case T_C_DOUBLE:
			sprintf(c->asString,"%f",dVal);
			break;
		case T_C_LONGLONG:
			sprintf(c->asString,"%f",(double)llVal/10000.0);
			break;
		case T_C_TIME:
			t=gmtime(&ttVal);

			sprintf(c->asString,"%04d%02d%02d%02d%02d%02d",
			  t->tm_year+1900,
			  t->tm_mon+1,
			  t->tm_mday,
			  t->tm_hour,
			  t->tm_min,
			  t->tm_sec);
			break;
		case T_C_STRING:
			strncpy(c->asString,sVal,256);
			c->asString[256]=0;
			break;
		}
	}

	return rc;
}

/*
	Environment change token

	+-----+-----+------+-------------+-------------+
	| tok | len | type | text1       | text2       |
	+-----+-----+------+-------------+-------------+
	  1     2     1      smallString   smallString

	type:
		0x01 - current database change
		0x04 - network fragment length change
 */
static int
handleEnvChange(MSSQL *s)
{
char buf1[4096];
char buf2[4096];
int type;
int rc;
int len;
	rc=read2Int(s,&len);
	rc|=read1Int(s,&type);
	if (s->verbose)
		fprintf(stderr,"Environment change: %02x: ",type);
	switch (type) {
	case 0x01:
		rc|=readSmallString(s,buf1,4095);
		rc|=readSmallString(s,buf2,4095);
		if (s->verbose)
			fprintf(stderr," current database: %s %s\n",
			  buf2,buf1);
		break;
	case 0x04:
		rc|=readSmallString(s,buf1,4095);
		rc|=readSmallString(s,buf2,4095);
		if (s->verbose)
			fprintf(stderr," network fragment: %s %s\n",
			  buf2,buf1);
		break;
	default:
		doRead(s,(unsigned char *)buf1,len-1);
		if (s->verbose)
			dump(buf1,len-1,"Unknown...");
		break;
	}
	return rc;
}

/*
	Login ACK

	+-----+-----+--------+----------+----------------+---------+
	| tok | len | status | protocol | server program | server  |
	|     |     |        | version  | name           | version |
	+-----+-----+--------+----------+----------------+---------+
	  1     2     1        4          smallString      4
*/
static int
handleLoginAck(MSSQL *s)
{
int status;
unsigned char protoVersion[4];
char progName[33];
unsigned char progVersion[4];
int rc;
int len;

	rc=read2Int(s,&len);
	rc|=read1Int(s,&status);
	rc|=doRead(s,protoVersion,4);
	rc|=readSmallString(s,progName,32);
	rc|=doRead(s,progVersion,4);
	
	if (s->verbose)
		fprintf(stderr,"Login ACK: status: %d, program name: %s\n",
		  status,progName);

	return rc;
}

/*
	Info - like Error
	
	+-----+-----+---------+-------+-------+---------+--------+-----+
	| tok | len | message | state | level | message | server | ??? |
	|     |     | number  |       |       | text    | name   |     |
	+-----+-----+---------+-------+-------+---------+--------+-----+
	  1     2     4         1       1       mediumS   smallS   3
*/
static int
handleInfo(MSSQL *s)
{
char buf[65];
int rc;
int len;
int message;
int state;
int level;
char buf1[3];

	rc=read2Int(s,&len);
	rc|=read4Int(s,&message);
	rc|=read1Int(s,&state);
	rc|=read1Int(s,&level);
	rc|=readMediumString(s,buf,64);
	rc|=readSmallString(s,0,1);
	rc|=doRead(s,buf1,3);
	if (s->verbose) {
		fprintf(stderr,"Info: %s,message=%d,state=%d,level=%d\n",
		  buf,message,state,level);
		dump(buf1,3,"");
	}
	return rc;
}


/*
	Error - like Info
	
	+-----+-----+---------+-------+-------+---------+--------+-----+
	| tok | len | message | state | level | message | server | ??? |
	|     |     | number  |       |       | text    | name   |     |
	+-----+-----+---------+-------+-------+---------+--------+-----+
	  1     2     4         1       1       mediumS   smallS   3
*/
static int
handleError(MSSQL *s)
{
char buf[65];
int rc;
int len;
int message;
int state;
int level;
char buf1[3];

	rc=read2Int(s,&len);
	rc|=read4Int(s,&message);
	rc|=read1Int(s,&state);
	rc|=read1Int(s,&level);
	rc|=readMediumString(s,buf,64);
	rc|=readSmallString(s,0,1);
	rc|=doRead(s,buf1,3);
	if (s->verbose) {
		fprintf(stderr,"Error: %s,message=%d,state=%d,level=%d\n",
		  buf,message,state,level);
		dump(buf1,3,"");
	}
	return rc;
}

/*
	Done

	+-----+--------+------+----------+
	| tok | status | info | affected |
	+-----+--------+------+----------+
	  1     2        2      4
*/
static int
handleDone(MSSQL *s)
{
int status;
int info;
int count;
int rc;

	rc=read2Int(s,&status);
	rc|=read2Int(s,&info);
	rc|=read4Int(s,&count);
	if (s->verbose)
		fprintf(stderr,"Done: status=%d,info=%d,count=%d\n",
		  status,info,count);
	return rc;
}

static int
handleToken(MSSQL *s,int token)
{
int rc=0;

	if (s->verbose)
		fprintf(stderr,"token=%02x\n",token);

	switch (token) {
	case TOKEN_COL_NAMES:
		rc=handleColumnNames(s);
		break;
	case TOKEN_COL_METADATA:
		rc=handleMetaData(s);
		break;
	case TOKEN_ENV_CHANGE:
		rc=handleEnvChange(s);
		break;
	case TOKEN_LOGIN_ACK:
		rc=handleLoginAck(s);
		break;
	case TOKEN_INFO:
		rc=handleInfo(s);
		break;
	case TOKEN_ERROR:
		rc=handleError(s);
		break;
	case TOKEN_DONE1:
	case TOKEN_DONE2:
	case TOKEN_DONE3:
		rc=handleDone(s);
		break;
	case TOKEN_ROW_DATA:
		rc=handleRowData(s);
		break;
	default:
		/* TDS error: unknown token! */
	}

	return rc;
}

int
doLogin(MSSQL *s,
  char *hostName,
  char *userName,
  char *password,
  char *id,
  char *appName,
  char *serviceName,
  char *language,
  char *charset
  )
{
unsigned char magic1[]= {
#if defined(HW_LITTLE_ENDIAN)
                           0x03,0x01,
#elif defined(HW_BIG_ENDIAN)
                           0x02,0x00,
#endif
                                     0x06,0x0a,0x09,0x01,0x01,0x00,0x00};
unsigned char magic2[]= {0x00,0x00,0x00};
unsigned char protocolVersion[]= {0x04,0x02,0x00,0x00};
unsigned char magic3[]= {0x00,13,17};
unsigned char magic4[]= {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};
unsigned char magic5=   0x01;
int rc;
int token;

	rc=writeString(s,hostName,30);         /* client host name */
	rc|=writeString(s,userName,30);        /* account name */
	rc|=writeString(s,password,30);        /* account password */
	rc|=writeString(s,id,30);              /* unique ID on client side ??? */
	rc|=writeBuf(s,magic1,9);
	rc|=write4Int(s,512);
	rc|=writeBuf(s,magic2,3);
	rc|=writeString(s,appName,30);
	rc|=writeString(s,serviceName,30);
	rc|=writeString(s,password,255);
	rc|=writeBuf(s,protocolVersion,4);     /* TDS version; { 0x04,0x02,0x00,0x00 } */
	rc|=writeString(s,"Pecioma",10);       /* client program name */
	rc|=write4Int(s,0);
	rc|=writeBuf(s,magic3,3);
	rc|=writeString(s,language,30);        /* language */
	rc|=writeBuf(s,magic4,14);
	rc|=writeString(s,charset,30);         /* charset */
	rc|=writeChar(s,magic5);
	rc|=writeString(s,"4096",6);           /* network fragment size */
	rc|=write4Int(s,0);
	rc|=write4Int(s,0);

	rc|=flushMsSql(s);

	while ((rc=read1Int(s,&token))==BUFFER_NOT_EMPTY) {
		if ((rc=handleToken(s,token))==BUFFER_UNDERRUN) {
			ERR("Underrun in Login()");
		}
	}

	resizeMsSql(s,4096);  /* set requested NW fragment size; see above!
				(should be done in the env. change handler) */
	setFlag(s,1);

	return rc;
}
	
/*
	Execute simple query, parse tokens up to and including the first result line

	return:
		ROW_READY - tuple was available
		DONE - no tuples available
		WAITING_ROW - somehing is wrong...
 */
int
execDirect(MSSQL *s,char *query)
{
int rc;
int token;

	writeBuf(s,query,strlen(query));
	flushMsSql(s);
	
	s->queryState=WAITING_ROW;

	while (s->queryState!=ROW_READY &&
	  (rc=read1Int(s,&token))==BUFFER_NOT_EMPTY) {
		if (handleToken(s,token)==BUFFER_UNDERRUN)
			break;
		switch (token) {
		case TOKEN_ROW_DATA:
			s->queryState=ROW_READY;
			break;
		case TOKEN_DONE1:
		case TOKEN_DONE2:
		case TOKEN_DONE3:
			s->queryState=DONE;
			break;
		}
	}

	return s->queryState;
}

/*
	Try to fetch next tuple if available, parse all other tokens also

	return:
		ROW_READY - tuple was available
		DONE - no tuples available
		WAITING_ROW - somehing is wrong...
 */
int
nextTuple(MSSQL *s)
{
int rc;
int token;

	if (s->queryState!=ROW_READY)
		return DONE;

	s->queryState=WAITING_ROW;

	while (s->queryState!=ROW_READY &&
	  (rc=read1Int(s,&token))==BUFFER_NOT_EMPTY) {
		if (handleToken(s,token)==BUFFER_UNDERRUN)
			break;
		switch (token) {
		case TOKEN_ROW_DATA:
			s->queryState=ROW_READY;
			break;
		case TOKEN_DONE1:
		case TOKEN_DONE2:
		case TOKEN_DONE3:
			s->queryState=DONE;
			break;
		}
	}

	return s->queryState;
}

void
printStats(MSSQL *s)
{
	fprintf(stderr,"Bytes in : %d\n",s->bytesIn);
	fprintf(stderr,"Bytes out: %d\n",s->bytesOut);
}

void
printColumnNames(MSSQL *s)
{
COLUMN *c;

	for (c=s->columns.next;c;c=c->next) {
		printf("%s, ",c->name);
	}
	printf("\n");
}

void
printResult(MSSQL *s)
{
COLUMN *c;

	for (c=s->columns.next;c;c=c->next) {
		printf("%s, ",c->asString);
	}
	printf("\n");
}
